import logging
from typing import Tuple

from xlsxwriter.workbook import Workbook, Worksheet

from excelipy.models import Style, Link
from excelipy.style import process_style
from excelipy.styles.link import DEFAULT_LINK_STYLE

log = logging.getLogger("excelipy")


def write_link(
        workbook: Workbook,
        worksheet: Worksheet,
        component: Link,
        default_style: Style,
        origin: Tuple[int, int] = (0, 0),
) -> Tuple[int, int]:
    log.debug(f"Writing link at {origin}")

    processed_style = process_style(
        workbook,
        [
            DEFAULT_LINK_STYLE,
            default_style,
            component.style,
        ],
    )

    col0, row0 = origin
    width = component.width
    height = component.height

    if component.merged and (width > 1 or height > 1):
        worksheet.merge_range(row0, col0, row0 + height - 1, col0 + width - 1, component.text, processed_style)
        worksheet.write_url(row0, col0, component.url, processed_style, component.text)
    else:
        for dy in range(height):
            for dx in range(width):
                row = row0 + dy
                col = col0 + dx
                if dy == 0 and dx == 0:
                    worksheet.write_url(row, col, component.url, processed_style, component.text)
                else:
                    worksheet.write_blank(row, col, "", processed_style)

    return width, height



import importlib.resources as pkg_resources
import io
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

import excelipy as ep
from test import resources


@pytest.fixture
def resources_path() -> Path:
    return Path(str(pkg_resources.files(resources)))


@pytest.fixture
def img_path(resources_path: Path) -> Path:
    return resources_path / "img.png"


@pytest.fixture
def sample_df() -> pd.DataFrame:
    return pd.DataFrame(
        {
            "testing": [1, 2, 3],
            "testing2": [1, 2, 3],
            "tested": ["Yay", "Thanks", "Bud"],
        }
    )


@pytest.fixture
def numeric_df() -> pd.DataFrame:
    return pd.DataFrame(
        {
            "integers": [0, 2, 3],
            "invalid": [1, 2, 3],
            "floats": [1.2, 2.3, 3.1],
            "big_numbers": [100000000, 2001230, np.inf],
            "percents": [0.2129, np.nan, 1.11],
        }
    )


@pytest.fixture
def empty_df() -> pd.DataFrame:
    num_rows = 2
    num_cols = 9
    cols = [" "] * num_cols
    return pd.DataFrame(" ", index=range(num_rows), columns=cols)


def test_api(
    sample_df: pd.DataFrame,
    empty_df: pd.DataFrame,
    img_path: Path,
    numeric_df: pd.DataFrame,
):
    tb = pd.DataFrame(
        {"testing": [ep.Link(text="google", url="https://www.google.com")]}
    )

    def get_match_style(result: int) -> ep.Style:
        return (
            ep.Style(
                background="#00ff1e",
                font_color="#ffffff",
                bold=True,
            )
            if result == 1
            else ep.Style(
                background="#ff0013",
                font_color="#ffffff",
                bold=True,
            )
        )

    @ep.row_wise
    def get_match_style_row(row) -> ep.Style:
        _, result, _ = row
        return (
            ep.Style(
                background="#00ff1e",
                font_color="#ffffff",
                bold=True,
            )
            if result == 1
            else ep.Style(
                background="#ff0013",
                font_color="#ffffff",
                bold=True,
            )
        )

    numeric_formats = {
        "integers": ".0f",
        "floats": ".2f",
        "big_numbers": ",.1f",
        "percents": ".1%",
        "invalid": "invalid",
    }

    style = ep.Style(background="#33c481")
    sheets = [
        ep.Sheet(
            name="Hello!",
            components=[
                ep.Text(
                    text="This is my table",
                    style=ep.Style(bold=True),
                    width=4,
                ),
                ep.Fill(),
                ep.Text(text="Monowidth"),
                ep.Table(data=empty_df),
                ep.Fill(
                    width=4,
                    style=ep.Style(background="#D0D0D0"),
                ),
                ep.Table(
                    data=sample_df,
                    header_style={
                        col: ep.Style(
                            bold=True,
                            border=5,
                            border_color="#F02932",
                        )
                        for col in sample_df.columns
                    },
                    body_style=ep.Style(font_size=18),
                    column_style={
                        "testing": ep.Style(
                            font_size=10,
                            align="center",
                            numeric_format=".0f",
                        ),
                        "testing2": get_match_style,
                    },
                    column_width={
                        "tested": 20,
                    },
                    row_style={
                        1: ep.Style(
                            border=2,
                            border_color="#F02932",
                        )
                    },
                    style=ep.Style(padding=1),
                ).with_stripes(pattern="even"),
                ep.Image(
                    path=img_path,
                    width=2,
                    height=5,
                    style=ep.Style(border=2),
                ),
                ep.Table(
                    data=numeric_df,
                    default_style=False,
                    header_filters=False,
                    column_style={
                        col: ep.Style(
                            numeric_format=numeric_formats.get(col),
                            align="center",
                            fill_inf="-",
                            fill_na="-",
                            fill_zero="-",
                        )
                        for col in numeric_df.columns
                    },
                ),
                ep.Text(text="Hello", width=10, style=style),
                ep.Text(text="Hello", width=10, style=style, merged=False),
                ep.Link(text="Hello", url="https://www.google.com"),
                ep.Link(text="Hello", url="https://www.google.com", width=2),
                ep.Link(
                    text="Hello", url="https://www.google.com", width=2, merged=False
                ),
                ep.Fill(width=10, style=style),
                ep.Fill(width=10, style=style, merged=False),
                ep.Table(data=tb),
                ep.Table(data=sample_df, idx_column_style={1: get_match_style_row}),
            ],
            style=ep.Style(
                font_size=14,
                font_family="Times New Roman",
                padding=1,
            ),
            grid_lines=False,
        ),
    ]

    excel = ep.Excel(path=io.BytesIO(), sheets=sheets)
    ep.save(excel)


def test_serialization_and_deserialization(sample_df: pd.DataFrame):
    json_lines = sample_df.to_json(orient="records", lines=True)
    dict_lines = sample_df.to_dict(orient="records")
    ep.Sheet.model_json_schema()
    ep.Table.model_validate({"data": json_lines})
    ep.Table.model_validate({"data": dict_lines})
    with pytest.raises(Exception):
        ep.Table.model_validate({"data": 1})
    ep.Table(data=sample_df).model_dump_json()

def test_load_ai_guide():
    assert isinstance(ep.AI_GUIDE, str)


if __name__ == "__main__":
    pytest.main([__file__])



__all__ = [
    "write_table",
    "write_fill",
    "write_text",
    "write_link",
    "write_image",
]

from excelipy.writers.fill import write_fill
from excelipy.writers.image import write_image
from excelipy.writers.link import write_link
from excelipy.writers.table import write_table
from excelipy.writers.text import write_text



import re
from importlib.resources import files
from pathlib import Path

from excelipy import resources

DATE_CONVERSION = {
    "%Y": "yyyy",
    "%y": "yy",
    "%m": "mm",
    "%d": "dd",
    "%H": "hh",
    "%M": "mm",
    "%S": "ss",
    "%b": "mmm",
    "%B": "mmmm",
}


def python_to_excel_fmt(fmt: str) -> str:
    fmt = (fmt or "General").strip()
    if not fmt or fmt == "General":
        return "General"

    # 1. Identify Python-specific patterns
    is_python_date = "%" in fmt and any(x in fmt for x in DATE_CONVERSION)

    # Python formats: .2f, ,.1f, .0%, .1%, %, f, d
    # Excel formats start with 0 or #: 0.0%, 0%, #,##0.00
    is_python_num = (
        bool(re.search(r"\.(\d+)[f%]|^[fd]$|^%$|^,\.\d+f$", fmt)) and not fmt[0] in "0#"
    )

    # 2. Handle Python Dates
    if is_python_date:
        for py, xl in DATE_CONVERSION.items():
            fmt = fmt.replace(py, xl)
        return fmt

    # 3. Handle Python Numerics
    if is_python_num:
        precision = re.search(r"\.(\d+)", fmt)
        decimals = int(precision.group(1)) if precision else 0
        decimal_part = f".{'0' * decimals}" if decimals > 0 else ""

        if "%" in fmt:
            return f"0{decimal_part}%"
        if "," in fmt:
            return f"#,##0{decimal_part}"
        return f"0{decimal_part}"  # Handles 'd' and 'f'

    # 4. If it's not a Python format, it's either already Excel or General
    return fmt


PROP_MAP = dict(
    align="align",
    valign="valign",
    font_size="font_size",
    font_color="font_color",
    font_family="font_name",
    bold="bold",
    border="border",
    border_left="left",
    border_right="right",
    border_top="top",
    border_bottom="bottom",
    border_color="border_color",
    background="bg_color",
    numeric_format="num_format",
    underline="underline",
)

PRE_PROCESS_MAP = dict(
    numeric_format=python_to_excel_fmt,
)

AI_GUIDE = (Path(str(files(resources))) / "AI.md").read_text(encoding="utf-8")



from excelipy.models import Style

DEFAULT_HEADER_STYLE = Style(
    bold=True,
    background="#ecedef",
    border=1,
    border_color="#9fa3ac",
)

DEFAULT_BODY_STYLE = Style(
    border=1,
    border_color="#9fa3ac",
)



import logging
from typing import Tuple

from xlsxwriter.workbook import Workbook, Worksheet

from excelipy.models import Fill, Style
from excelipy.style import process_style

log = logging.getLogger("excelipy")


def write_fill(
        workbook: Workbook,
        worksheet: Worksheet,
        component: Fill,
        default_style: Style,
        origin: Tuple[int, int] = (0, 0),
) -> Tuple[int, int]:
    log.debug(f"Writing fill at {origin}")

    style = process_style(workbook, [default_style, component.style])
    col0, row0 = origin
    width = component.width
    height = component.height

    if component.merged and (width > 1 or height > 1):
        worksheet.merge_range(row0, col0, row0 + height - 1, col0 + width - 1, "", style)
    else:
        for dy in range(height):
            for dx in range(width):
                row = row0 + dy
                col = col0 + dx
                worksheet.write_blank(row, col, "", style)

    return width, height



__all__ = [
    "Style",
    "Component",
    "Fill",
    "Image",
    "Text",
    "Link",
    "Table",
    "Group",
    "Sheet",
    "Excel",
    "save",
    "row_wise",
    "unnest_components",
    "AI_GUIDE",
]

from excelipy.const import AI_GUIDE
from excelipy.models import (
    Style,
    Component,
    Fill,
    Image,
    Text,
    Table,
    Sheet,
    Excel,
    Link,
    Group,
)

from excelipy.service import save, unnest_components
from excelipy.writers.table import row_wise



import logging
from collections import defaultdict
from functools import wraps
from typing import Tuple, Dict, Optional

import numpy as np
import pandas as pd
from PIL import ImageFont
from xlsxwriter.workbook import Workbook, Worksheet

from excelipy.models import Style, Table, Link
from excelipy.style import process_style
from excelipy.styles.table import DEFAULT_HEADER_STYLE, DEFAULT_BODY_STYLE

log = logging.getLogger("excelipy")

DEFAULT_FONT_SIZE = 11
ROW_WISE_ARG = "_ep_row_wise"


def row_wise(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    setattr(wrapper, ROW_WISE_ARG, True)
    return wrapper


def _static_col_style(component: Table, col_name: str, col_idx: int) -> Style:
    maybe = component.idx_column_style.get(col_idx) or component.column_style.get(
        col_name
    )
    return Style() if callable(maybe) or maybe is None else maybe


def get_text_size(
    text: str,
    font_size: Optional[int] = None,
    font_family: Optional[str] = None,
):
    text = str(text)
    cur_font_size = font_size or DEFAULT_FONT_SIZE
    cur_font_family = font_family or "Arial"

    try:
        cur_font = ImageFont.truetype(
            f"{cur_font_family}.ttf".lower(),
            cur_font_size,
        )
    except Exception as e:
        cur_font = ImageFont.load_default()
        log.debug(
            f"Could not load custom font {cur_font_family}, using default. Exception: {e}",
        )

    return cur_font.getlength(text)


def get_style_font_family(*styles: Style) -> Optional[str]:
    cur_font = None
    for s in filter(None, styles):
        cur_font = s.font_family or cur_font
    return cur_font


def get_style_font_size(*styles: Style) -> Optional[int]:
    cur_font = None
    for s in filter(None, styles):
        cur_font = s.font_size or cur_font
    return cur_font


def get_auto_width(
    cur_col: str,
    col_idx: int,
    data: pd.Series,
    component: Table,
    cur_skip: bool,
    default_style: Style,
) -> float:
    header_font_size = get_style_font_size(
        DEFAULT_HEADER_STYLE,
        component.style,
        component.header_style.get(cur_col),
    )
    header_font_family = get_style_font_family(
        DEFAULT_HEADER_STYLE,
        component.style,
        component.header_style.get(cur_col),
    )
    header_len = get_text_size(
        cur_col,
        header_font_size,
        header_font_family,
    )
    cells_for_header = len([it for it in component.data.columns if it == cur_col])

    if cur_skip:
        header_len /= cells_for_header

    col_font_size = get_style_font_size(
        default_style,
        component.style,
        component.body_style,
        _static_col_style(component, cur_col, col_idx),
    )
    col_font_family = get_style_font_family(
        default_style,
        component.style,
        component.body_style,
        _static_col_style(component, cur_col, col_idx),
    )
    all_col_len = data.apply(str).apply(
        lambda it: get_text_size(
            it,
            col_font_size,
            col_font_family,
        ),
    )
    col_len = all_col_len.max()
    max_len = max(header_len, col_len)
    result = max_len // component.auto_width_tuning + component.auto_width_padding
    return result


def write_table(
    workbook: Workbook,
    worksheet: Worksheet,
    component: Table,
    default_style: Style,
    origin: Tuple[int, int] = (0, 0),
) -> Tuple[int, int]:
    x_size = component.data.shape[1]
    y_size = component.data.shape[0] + 1  # +1 for the header row

    col_size_cache_by_sheet = getattr(workbook, "_excelipy_col_size_cache", None)
    if col_size_cache_by_sheet is None:
        col_size_cache_by_sheet: Dict[str, Dict[int, int]] = {}
        setattr(workbook, "_excelipy_col_size_cache", col_size_cache_by_sheet)

    idx_by_header = defaultdict(list)
    if component.merge_equal_headers:
        for idx, cur_col in enumerate(component.data.columns):
            past_idx = idx_by_header.get(cur_col, [idx - 1])[-1]
            if idx == past_idx + 1:
                idx_by_header[cur_col].append(idx)
        idx_by_header = {k: v for k, v in idx_by_header.items() if len(v) >= 2}

    for col_idx, cur_col in enumerate(component.data.columns):
        current_header_style = component.header_style.get(
            cur_col,
            component.style,
        )
        header_styles = [default_style, current_header_style]

        if component.default_style:
            header_styles = [DEFAULT_HEADER_STYLE] + header_styles

        header_format = process_style(workbook, header_styles)
        header_write_skip = idx_by_header.get(cur_col, [])
        is_first = False
        cur_skip = col_idx in header_write_skip and not (
            is_first := col_idx == header_write_skip[0]
        )

        if is_first:
            merge_size = len(header_write_skip)
            worksheet.merge_range(
                origin[1],
                origin[0],
                origin[1],
                origin[0] + merge_size - 1,
                "",
                header_format,
            )

        worksheet.write(
            origin[1],
            origin[0] + col_idx,
            cur_col if not cur_skip else "",
            header_format,
        )

        set_width = component.idx_column_width.get(
            col_idx
        ) or component.column_width.get(cur_col)
        if set_width:
            estimated_width = set_width
        else:
            data = component.data.iloc[:, col_idx]
            estimated_width = get_auto_width(
                cur_col,
                col_idx,
                data,
                component,
                cur_skip,
                default_style,
            )
            col_size_cache = col_size_cache_by_sheet.get(worksheet.name, {})
            cur_cached = col_size_cache.get(origin[0] + col_idx, 0)
            estimated_width = max((cur_cached, estimated_width))
            col_size_cache[origin[0] + col_idx] = estimated_width
            col_size_cache_by_sheet[worksheet.name] = col_size_cache

        log.debug(
            f"Estimated width for {cur_col}: {estimated_width} [Sheet: {worksheet.name}]",
        )
        worksheet.set_column(
            origin[0] + col_idx,
            origin[0] + col_idx,
            int(estimated_width),
        )

    if component.header_filters:
        worksheet.autofilter(
            origin[1],
            origin[0],
            origin[1],
            origin[0] + len(list(component.data.columns)) - 1,
        )

    for col_idx, col in enumerate(component.data.columns):
        col_style = _static_col_style(component, col, col_idx)
        for row_idx, (_, row) in enumerate(component.data.iterrows()):
            row_style = component.row_style.get(row_idx)
            body_style = [
                default_style,
                component.style,
                component.body_style,
                col_style,
                row_style,
            ]

            if component.default_style:
                body_style = [DEFAULT_BODY_STYLE] + body_style

            merged_style = Style()
            body_style = list(filter(None, body_style))
            for style in body_style:
                merged_style = merged_style.merge(style)

            cell = row.iloc[col_idx]
            url = None

            if isinstance(cell, Link):
                url = cell.url
                cell = cell.text

            if merged_style.fill_na is not None and pd.isna(cell):
                cell = merged_style.fill_na
                merged_style = merged_style.model_copy(update=dict(numeric_format=None))

            if merged_style.fill_zero is not None and cell == 0:
                cell = merged_style.fill_zero
                merged_style = merged_style.model_copy(update=dict(numeric_format=None))

            if merged_style.fill_inf is not None and cell in (np.inf, -np.inf):
                cell = merged_style.fill_inf
                merged_style = merged_style.model_copy(update=dict(numeric_format=None))

            maybe_callable = component.idx_column_style.get(col_idx) or component.column_style.get(col)

            if callable(maybe_callable):
                if getattr(maybe_callable, ROW_WISE_ARG, False):
                    dyn_style = maybe_callable(row)
                else:
                    dyn_style = maybe_callable(cell)

                if dyn_style is not None:
                    merged_style = merged_style.merge(dyn_style)

            if row_style is not None:
                merged_style = merged_style.merge(row_style)

            current_format = process_style(workbook, [merged_style])
            if url is None:
                worksheet.write(
                    origin[1] + row_idx + 1,
                    origin[0] + col_idx,
                    cell,
                    current_format,
                )
            else:
                worksheet.write_url(
                    origin[1] + row_idx + 1,
                    origin[0] + col_idx,
                    url,
                    current_format,
                    cell,
                )

    return x_size, y_size
